/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/search.c
    Version 1.0
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include "gn.h"
#include "search.h"
#include "reg.h"
#include "regi.h"

#define	ERROR	(-1)
#define MATCH	(0)
#define NOMATCH	(1)


/* make up for deficient stat.h's */

#ifndef S_ISDIR
#define S_ISDIR(mode) ((mode) & S_IFREG)
#endif


static int	found_match();
static void	dogrep(),
		www_results(),
		range_search();

/*
 * sendsearch( ip)  Send the items from a "virtual cachefile" produced
 * running the search program ip->filepath.
 */

void
sendsearch(  ip)
Item	*ip;

{
	FILE	*fp;
	char	linebuf[PATHLEN];
	int	qflag = FALSE;

	struct stat stat_buf;

	if ( *(ip->extra) == '\0') {
		if ( iswww )
			qflag = TRUE;
		else {
			senderr("Search string is empty!", ip);
			exit( 2);
		}
	}

	switch( ip->gtype[1]) {
		case 'c':
			if ( qflag)
				www_query( ip, KEYMSG);
			else
				cache_search( ip);
			return;

		case 'm':
			if ( qflag)
				www_query( ip, REGMSG);
			else
				range_search( ip);
			return;

		case 'g':
			if ( qflag)
				www_query( ip, REGMSG);
			else
				dogrep( ip);
			return;
		case '\0':
			if ( (stat( ip->filepath, &stat_buf) == 0) && 
					(S_ISDIR( stat_buf.st_mode)) ) {
					/* It's a grep search */
				if ( qflag)
					www_query( ip, REGMSG);
				else
					dogrep( ip);
				return;
			}
	}
	
	if ( qflag) {
		www_query( ip, ASKWORD);
		return;
	}
	if ( (fp = safer_popen( ip->filepath, ip->extra, "r"))
					 == (FILE *) NULL ) {
		senderr2( "Can't exec", ip->filepath);
		exit( 2);
	}

	sprintf(linebuf, "%s `%s':", RESPNSE, ip->extra);
	www_results( ip, "Response", linebuf);


	while ( get_cache_line( linebuf, fp)) {
		if ( iswww)
			www_cache_line( linebuf);
		else
			send_cache_line( linebuf);
	}	
	pclose( fp);
	cache_epilog( iswww);
	writelog( "Search", ip->selector);
}

static void
www_results(ip, str1, str2)
Item	*ip;
char	*str1,
	*str2;
{

	char	*cp,
		buf[BUFSIZE];

	if ( !iswww)
		return;
	cp = ( (*(ip->name)) ? ip->name : str1 );

	send_text_line( "<ISINDEX>" );
	sprintf(buf,"<title> %s </title><h2> %s</h2>", cp, cp);
	send_text_line( buf);
	send_text_line( str2);
	send_text_line( "<p><ul>" );
}	


void
www_query( ip, str)
Item	*ip;
char	*str;
{
	char	*cp,
		buf[BUFSIZE];

	cp = ( (*(ip->name)) ? ip->name : str );

	sprintf(buf,"<ISINDEX><title> %s </title><h2> %s </h2>", cp, cp);
	send_text_line( buf);
	strcpy( buf, str);
	send_text_line( buf);
}






/*
 * range_search( ip)  This is type "7m",  we open the composite cachefile
 * basename..cache and also the composite file "basename".  For each
 * cachefile line search the corresponding range in the composite file
 * for lines matching the search term pointed to by ip->regptr.
 * When a match is found send the cachefile line.
 */

static void
range_search( ip)
Item	*ip;

{
	Item		curitem,
			*item_p;

	Cache_entry	cstruct,
			*cstructp;
	FILE	*fp,
		*cfp;

	struct regprog	*regp;

	char	cachepath[PATHLEN],
		cline[BUFSIZE],
		linebuf[BUFSIZE];

	item_p = &curitem;
	cstructp = &cstruct;

	if ( (regp = regcomp( strlower( ip->extra))) == NULL ) {
		senderr2( "Search term regular expression error: ",
			ip->extra);
		exit( 2);
	}


	sprintf( cachepath, "%s.%s", ip->filepath, cfname);

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", ip->filepath);
		exit( 2);
	}
	if ( (cfp = fopen( cachepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", cachepath);
		exit( 2);
	}

	sprintf( linebuf, "%s `<b>%s</b>'.  %s", REGRES1, ip->extra, REGRES2);
	www_results( ip, "Structured File Search", linebuf);

	while ( get_cache_line( cline, cfp)) {
		parse_cache( cline, cstructp);
		parse_selector( item_p, cstructp->path, FULL);

		fseek( fp, item_p->range_start, 0);
		while ( fgets( linebuf, BUFSIZE, fp) && 
				(ftell( fp) <= item_p->range_end)) {
			chop( linebuf);
			if ( regfind( regp, strlower( linebuf)) ) {
				if ( iswww )
					www_cache_line( cline);
				else
					send_cache_line( cline);
				break;
			}
		}
	}	
	cache_epilog( iswww);
	writelog( "Sent range search", ip->selector);
}







/* Do regexp search of directory pointed to by ip->filepath */

static void
dogrep( ip)
Item	*ip;
{
	FILE	*fp,
		*gfp;

	struct regprog	*regp;

	char	*cp,
		cachefile[PATHLEN],
		tmpbuf[PATHLEN],
		pathbuf[PATHLEN],
		linebuf[PATHLEN];

	
	if ( streq( ip->gtype, "7g") )
		strcpy(cachefile, ip->filepath);
	else
		sprintf( cachefile, "%s/%s", ip->filepath, cfname);

	if ( (fp = fopen( cachefile, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open", cachefile);
		exit( 2);
	}

	sprintf( tmpbuf, "%s `<b>%s</b>'.  %s", REGRES1, ip->extra, REGRES2);
	www_results( ip, "Directory Regular Expression Search", tmpbuf);

	while ( fgets( linebuf, BUFSIZE, fp)) {
		strcpy( tmpbuf, rootdir);
		cp = strchr( linebuf, '\t');
		if ( *++cp != '0' )	/* Quit if not type 0 (file) */
			continue;
		strcat( tmpbuf, ++cp);
		cp = strchr( tmpbuf, '\t'); 
		*cp = '\0';
		if ( (cp = strrchr( tmpbuf, '(')) != (char *) NULL)
			*cp = '\0';
		cp = strchr( tmpbuf, '/');
		sprintf( pathbuf, "%s", cp);

                if ( (gfp = fopen( pathbuf, "r")) == (FILE *) NULL ) {
                        /* File is no longer there, keep going */
                        continue;
                }

		if ( (regp = regcomp( strlower(ip->extra))) == NULL ) {
			senderr2( "Search term regular expression error: ",
					ip->extra);
			break;
		}		
		if ( found_match( regp, gfp) )
			if ( iswww)
				www_cache_line( linebuf);
			else
				send_cache_line( linebuf);

		fclose( gfp);
	}
	cache_epilog( iswww);
}


static int
found_match( rp, fp)
struct regprog	*rp;
FILE	*fp;
{
	char linebuf[BUFSIZE];

	while ( fgets( linebuf, BUFSIZE, fp)) {
		strlower( linebuf);
		if ( regfind( rp, linebuf) )
			return TRUE;
	}
	return FALSE;
}

