/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/sio.c
    Version 1.0
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include "gn.h"

static int	isgnlink = FALSE;
static void	www_escape();

void
get_selector( sel)
char	*sel;
{
	fgets( sel, PATHLEN, stdin);
	chop( sel);
}


void
send_text_line( line)
char	*line;
{
	chop( line);
	if ( streq( line, "."))
	strcat( line, ".");
		/* Double a '.' if on line by itself */
	fputs( line, stdout);
	fputs( "\r\n", stdout);
}

void
send_cache_line( line)
char	*line;
{
	char	*cp,
		*cp2;
	int	brackdepth = 0;

	Item    item;
	Cache_entry cache_entry;

	if ( strncmp( line, "http:", 5) == 0 )
		return;

	if ( accesstype == DIRCHK) {
		parse_cache(line, &cache_entry);
		parse_selector( &item, cache_entry.path, FULL);
		if (!chkaccess( &item, NOLOG))  
			return;
	}
	if ( *line == 'i') {  /* Delete everything in <> angle brackets */
		cp = cp2 = line;
		while ( *cp2 ) {
			if ( *cp2 == '<')
				brackdepth++;
			if ( brackdepth >0 )
				cp2++;
			else 
				*cp++ = *cp2++;
			if ( *(cp2 - 1) == '>')
				brackdepth--;
		}
		*cp = '\0';
	}
	fputs( line, stdout);
	fputs( "\r\n", stdout);
}

void
send_end()
{
	fputs( ".\r\n", stdout);
}

void
www_cache_line( line)
char	*line;
{
	register char	*cp,
			*cp2;
	char		buf[PATHLEN];

	Item    	item;
	Cache_entry	cache_entry,
			*chp;
	char	out[PATHLEN];

	if ( strncmp( line, "http:", 5) == 0 ) {
		strcpy( out, line+5);
		send_text_line( out);
		return;
	}
	chp = &cache_entry;
	parse_cache(line, chp);

	if ( accesstype == DIRCHK) {
		strcpy( buf, chp->path);
		parse_selector( &item, buf, FULL);
		if (!chkaccess( &item, NOLOG))  
			return;
	}

	cp = chp->path;

	www_escape( chp->path);
	switch (*line) {
	case '0':
	case '1':
	case '7':
		if ( isgnlink) {
			if ( (strncmp( chp->path, "ftp", 3) == 0 ) &&
				((cp2 =  strchr(chp->path, ':')) != NULL) &&
				((cp =  strchr(cp2, '@')) != NULL) ) {

				*cp++ = '\0';
				cp2++;

				sprintf( out, 
				"<li> <a href=\"ftp://%s%s\">%s</a>\r\n", 
					cp2, cp, chp->name);
			}
			else
				sprintf( out, 
				"<li> <a href=\"http://%s:%s/%s\">%s</a>\r\n", 
				chp->host, chp->port, chp->path, chp->name);
			break;
		}

	case '9':
	case 'I':
	case 's':
		sprintf(out,"<li> <a href=\"gopher://%s:%s/%c%s\">%s</a>\r\n", 
			chp->host, chp->port, *line, chp->path, chp->name);
		break;
	case 'i':
		strcpy(out, ++line);
		cp = strchr( out, '\t');
		*cp = '\0';
		break;
	}
	send_text_line( out);
}




/*
 * www_escape( path) replaces special chars in path with %XX escapes
 */

static void
www_escape( path)
char	*path;
{

	register char	*cp,
			*cp2;
	char	buf[BUFSIZE];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		switch (*cp) {
		case '"':
		case ' ':
			sprintf( cp2, "%%%X", (int) *cp);
			cp2 += 3;
			cp++;
			break;
		default:
			*cp2++ = *cp++;
		}
	}
	*cp2 = '\0';
	strcpy( path, buf);
}

/*
 * www_unescape( path) undoes what www_escape does
 */

void
www_unescape( path)
char	*path;
{

	register char	*cp,
			*cp2;
	int	val;

	char	minbuf[3],
		buf[BUFSIZE];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		if ( *cp == '%') {
			cp++;
			minbuf[0] = *cp++;
			minbuf[1] = *cp++;
			minbuf[3] = '\0';
			sscanf( minbuf, "%x", &val);
			*cp2++ = (char) val;
		}
		else
			*cp2++ = *cp++;
	}
	*cp2 = '\0';
	strcpy( path, buf);
}


char *
get_cache_line( buf, fp)
char	*buf;
FILE	*fp;
{
	char	*cp;

	isgnlink = TRUE;
	if (  (cp = fgets( buf, BUFSIZE, fp)) == NULL)
		return NULL;
	while (*cp)
		cp++;
	
	cp -= 2;
	if (*cp == '\r')
		cp--;
	if (*cp == '*') {
		*cp = '\0';
		isgnlink = FALSE;
	}
	else
		*++cp = '\0';
	return buf;
}
		
