/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/csearch.c
    Version 1.13
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include "gn.h"
#include "reg.h"
#include "search.h"


static int	nomatch = TRUE;
static void	csearch();


static struct regprog	*regp;

static Cache_entry	cache_line;



void
cache_search( ip)
Item	*ip;
{

	int	depth = MAXDEPTH;

	char	buf[BUFSIZE];

	if ( http) {
		sprintf(buf, "Document Title Search");
	}

	cache_prolog( ip, buf, "<ISINDEX>");

	sprintf( buf, "^[^\t]*%s", ip->extra);
	if ( (regp = regcomp( strlower( buf))) == NULL ) {
		senderr2( "Search term regular expression error: ",
			ip->extra);
		exit( 2);
	}

	if ( http) {
		sprintf(buf, 
		"</ul>Matches for the regular expression <b>`%s'</b>:<ul>",
				ip->extra);
		send_text_line( buf);
	}
	csearch( ip->filepath, depth);
	if ( http && nomatch) {
		sprintf( buf, "</ul><p>%s `%s'. %s<p>",
			REGMISS, ip->extra, REGRES2);
		send_text_line( buf);
	}
	else if (http) {
		sprintf( buf, "</ul><p> %s <p>", REGRES2);
		send_text_line( buf);
	}
	cache_epilog( http);
	writelog( "Sent cache search", ip->selector);
}	

/*
 * csearch( cpath, depth)
 * Search cache file with path cpath for lines which match the compiled
 * regular expression pointed to by static pointer regp.  When found
 * call send_cache_line. If  a cachefile
 * line corresponds to a directory, make a new cpath for this directory
 * and recursively call csearch for that cachefile.  Note: with symbolic
 * links there could be an infinite loop of "subdirectories" -- to handle
 * this problem, go only depth levels deep in the recursion.  This will
 * still result in a virtual directory with many duplicates.
 */

static void
csearch( cpath, depth)
char	*cpath;
int	depth;
{
	FILE	*cfp;

	struct stat stat_buf;

	char	*cp,
		*cbuf,
		*cline2,
		cbuf2[MAXLEN],
		cnewpath[MAXLEN];

	if ( depth <= 0 )
		return;

	if ( (cfp = fopen( cpath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open cachefile (csearch): ", cpath);
		exit( 2);
	}

	while ( read_cache( &cbuf, &cline2, cfp)) {
		if ( *cbuf =='i' || *cbuf == '\t' ||
			( strncmp( cbuf, "http:", 5) == 0 ))
			continue;
	/* Don't search menu comments (type 'i' or "http:" or 2nd line  */

		strcpy( cbuf2, cbuf);
		if ( (cp = strchr( cbuf2, '\t')) == NULL ) {
			senderr2( "Corrupt cache entry: ", cbuf);
			exit( 2);
		}
		*cp = '\0';
		
		if ( regfind( regp, strlower( cbuf2 + 1)) )
			if ( http) {
				nomatch = FALSE;
				www_cache_line( cbuf);
			}
			else
				send_cache_line( cbuf);

		if ( *cbuf == '1' && cline2 && *cline2) {
		/* It's a directory and *cline2 != 0, so it's not remote */
		/* Build new menu and cache path names and recurse */

			parse_cache( cbuf, &cache_line);
			if ( (cp = strchr( cache_line.path, '/')) == NULL) {
				senderr2( "Corrupt cache entry: ", cbuf);
				exit( 2);
			}

			strcpy( cnewpath, rootdir);
			strcat( cnewpath, cp);
			if ( streq( cache_line.path, rootdir))
	        	        continue;  /* Never go back to root */
			if ( cache_line.path[1] == 'm' ) {
	        	        strcat( cnewpath, ".");
	        	        strcat( cnewpath, cfname);
			}
			else {
	        	        strcat( cnewpath, "/");
	        	        strcat( cnewpath, cfname);
			}
			if ( stat( cnewpath, &stat_buf) == 0 )
				csearch( cnewpath, depth - 1);
		}
	}
	fclose( cfp);
}


/*
 * parse_cache( line, cip) looks at a line from a .cache file and fills in
 * the fields of the Cache_entry struct pointed to by cip.  These fields
 * type1, name,  path, host and port.  The path has ending '/' removed 
 * unless it is the root which  could have path / or 1/ or 7c/.  
 * Type1 is the single character type which is the first character of
 * the .cache line.
 */


void
parse_cache( line, cip)
char		*line;
Cache_entry	*cip;
{
	char	*cp,
		*cp2,
		*cp3,
		*cp4,
		cline[MAXLEN];


	assert ( (*line != '\t') );
	strcpy( cline, line);
	chop( cline);
	if ( (cp = strchr( cline, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no tabs): ", cline);
		exit( 2);
	}

	*cp++ = '\0';
	cip->type1 = cline[0];
	strcpy( cip->name, &cline[1]);
	
	/* cp points after 1st tab */

	if ( (cp2 = strchr( cp, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no 2nd tab): ", cline);
		exit( 2);
	}
	*cp2++ = '\0';

	/* cp2 points after 2nd tab */

	if ( (*cp == '\0') || streq( cp, "1"))
		strcpy(cip->path, "1/");
	else
		strcpy( cip->path, cp);

	/* Make sure path does not end in '/' except for root */

	if (cip->path[strlen(cip->path)-1] == '/')
		cip->path[strlen(cip->path)-1] = '\0';

	if ( (cp3 = strrchr( cip->path, '/')) == NULL)	/* It's the root */
		strcat( cip->path, "/");  /* Put back '/' */

	if ( (cp3 = strchr( cp2, '\t')) == (char *) NULL) {
		senderr2( "Corrupted cache entry (no 3rd tab): ", cline);
		exit( 2);
	}
	*cp3++ = '\0';

	/* cp3 points after 3rd tab */

	strcpy( cip->host, cp2);
	if ( (cp4 = strchr( cp3, '\t')) !=  NULL)
		*cp4 = '\0';
	strcpy( cip->port,  cp3);

}




