/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/search.c
    Version 1.1
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <sys/types.h>
#include <sys/stat.h>
#include <stdio.h>
#include <string.h>
#include "gn.h"
#include "search.h"
#include "reg.h"
#include "regi.h"

#define	ERROR	(-1)
#define MATCH	(0)
#define NOMATCH	(1)


/* make up for deficient stat.h's */

#ifndef S_ISDIR
#define S_ISDIR(mode) ((mode) & S_IFREG)
#endif


static int	found_match();
static void	dogrep(),
		dowais(),
		www_results(),
		range_search();

/*
 * sendsearch( ip)  Send the items from a "virtual cachefile" produced
 * running the search program ip->filepath.
 */

void
sendsearch(  ip)
Item	*ip;

{
	FILE	*fp;
	char	linebuf[PATHLEN],
		*line1,
		*dummy;
	int	qflag = FALSE;

	if ( *(ip->extra) == '\0') {
		if ( http )
			qflag = TRUE;
		else {
			senderr("Search string is empty!", ip);
			exit( 2);
		}
	}

	switch( ip->gtype[1]) {
		case 'c':
			if ( qflag)
				www_query( ip, KEYMSG);
			else
				cache_search( ip);
			return;

		case 'm':
			if ( qflag)
				www_query( ip, REGMSG);
			else
				range_search( ip);
			return;

		case 'g':
			if ( qflag)
				www_query( ip, REGMSG);
			else
				dogrep( ip);
			return;

		case 'w':
			if ( qflag)
				www_query( ip, WAISMSG);
			else
				dowais( ip);
			return;

		case '\0':
			break;

		default:
			senderr2( "Unknown selector type %s", ip->gtype);
	}
	
	if ( qflag) {
		www_query( ip, ASKWORD);
		return;
	}

	if ( (fp = safer_popen( ip->filepath, ip->args, "r"))
					 == (FILE *) NULL ) {
		senderr2( "Can't exec", ip->filepath);
		exit( 2);
	}

	sprintf(linebuf, "%s <b>`%s'</b>:", RESPNSE, ip->extra);
	www_results( ip, "Response", linebuf);


	while ( read_cache( &line1, &dummy, fp)) {
		if ( http)
			www_cache_line( line1);
		else
			send_cache_line( line1);
	}	
	pclose( fp);
	cache_epilog( http);
	writelog( "Search", ip->selector);
}

static void
www_results(ip, str1, str2)
Item	*ip;
char	*str1,
	*str2;
{

	char	*cp,
		buf[BUFSIZE];

	if ( !http)
		return;
	http_prolog( ip);
	cp = ( (*(ip->name)) ? ip->name : str1 );

	send_text_line( "<HEAD><ISINDEX>" );
	sprintf(buf,"<TITLE> %s </TITLE></HEAD><BODY><H2> %s</H2>", cp, cp);
	send_text_line( buf);
	send_text_line( str2);
	send_text_line( "<P><UL>" );
}	


void
www_query( ip, str)
Item	*ip;
char	*str;
{
	char	*cp,
		buf[BUFSIZE];

	http_prolog( ip);
	cp = ( (*(ip->name)) ? ip->name : str );

	sprintf(buf, "<HEAD><TITLE> %s </TITLE></HEAD>", cp);
	send_text_line( buf);
	sprintf(buf, "<BODY><H2> %s </H2>", cp);
	send_text_line( buf);
	strcpy( buf, str);
	send_text_line( buf);
	send_text_line( "<P><ISINDEX></BODY>" );
}






/*
 * range_search( ip)  This is type "7m",  we open the composite cachefile
 * basename..cache and also the composite file "basename".  For each
 * cachefile line search the corresponding range in the composite file
 * for lines matching the search term pointed to by ip->regptr.
 * When a match is found send the cachefile line.
 */

static void
range_search( ip)
Item	*ip;

{
	Item		curitem,
			*item_p;

	Cache_entry	cstruct,
			*cstructp;
	FILE	*fp,
		*cfp;

	struct regprog	*regp;

	int	firsttime = TRUE;

	char	*cline,
		*dummy,
		cachepath[PATHLEN],
		msgbuf[BUFSIZE],
		linebuf[BUFSIZE];

	item_p = &curitem;
	cstructp = &cstruct;

	if ( (regp = regcomp( strlower( ip->extra))) == NULL ) {
		senderr2( "Search term regular expression error: ",
			ip->extra);
		exit( 2);
	}


	sprintf( cachepath, "%s.%s", ip->filepath, cfname);

	if ( (fp = fopen( ip->filepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", ip->filepath);
		exit( 2);
	}
	if ( (cfp = fopen( cachepath, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open file", cachepath);
		exit( 2);
	}

	sprintf( msgbuf, "%s `<b>%s</b>'.  %s", REGRES1, ip->extra, REGRES2);

	while ( read_cache( &cline, &dummy, cfp)) {
		parse_cache( cline, cstructp);
		parse_selector( item_p, cstructp->path, FULL);

		fseek( fp, item_p->range_start, 0);
		while ( fgets( linebuf, BUFSIZE, fp) && 
				(ftell( fp) <= item_p->range_end)) {
			chop( linebuf);
			if ( regfind( regp, strlower( linebuf)) ) {
				if ( http ) {
					if (firsttime == TRUE) {
						firsttime = FALSE;
						www_results( ip,
						"Structured File Search",
							msgbuf);
					}
					www_cache_line( cline);
				}
				else
					send_cache_line( cline);
				break;
			}
		}
	}	
	if ( firsttime == TRUE) {
		sprintf( msgbuf, "%s <b>`%s'</b>.  %s", REGMISS,
			ip->extra, REGRES2);
		www_results( ip, "Structured File Search", msgbuf);
	}
	cache_epilog( http);
	writelog( "Sent range search", ip->selector);
}


/* Do WAIS search of index in directory  ip->filepath */

static void
dowais( ip)
Item	*ip;
{
	char	*title,
		*protocol,
		waisgnfile[PATHLEN];

	strcpy( waisgnfile, WAISGN);
	if ( *waisgnfile == '\0')
		senderr( "Waisgn is not installed", ip);
	protocol = ( http ? "http" : "gopher");

	http_prolog( ip);
	title = ( (*(ip->name)) ? ip->name : "WAIS Index Search Results" );
	
	writelog( "Sent WAIS search", ip->selector);
	execl( waisgnfile, "waisgn", ip->filepath, rootdir, cfname, myhost,
		myport, ip->gtype, title, protocol, ip->extra, (char *) NULL);
	senderr2( "Couldn't exec waisgn", waisgnfile);
}

/* Do regexp search of directory pointed to by ip->filepath */

static void
dogrep( ip)
Item	*ip;
{
	FILE	*fp,
		*gfp;

	struct regprog	*regp;

	register char	*cp;
	char	*line1,
		*dummy,
		cachefile[PATHLEN],
		tmpbuf[PATHLEN],
		msgbuf[BUFSIZE];

	int	is0h = FALSE,
		firsttime = TRUE;
	
	if ( streq( ip->gtype, "7g") )
		strcpy(cachefile, ip->filepath);
	else
		sprintf( cachefile, "%s/%s", ip->filepath, cfname);

	if ( (fp = fopen( cachefile, "r")) == (FILE *) NULL ) {
		senderr2( "Can't open", cachefile);
		exit( 2);
	}

	sprintf( msgbuf, "%s <b>`%s'</b>.  %s", REGRES1, ip->extra, REGRES2);
	while ( read_cache( &line1, &dummy, fp)) {
		strcpy( tmpbuf, rootdir);
		cp = strchr( line1, '\t');
		if ( *++cp != '0' )	/* Quit if not type 0 (file) */
			continue;
		if ( *++cp == 'h') {
			is0h = TRUE;
			cp++;
		}
		strcat( tmpbuf, cp);
		cp = strchr( tmpbuf, '\t'); 
		*cp = '\0';
		if ( (cp = strrchr( tmpbuf, '(')) != (char *) NULL)
			*cp = '\0';
		if ( is0h ) {
			strcat ( tmpbuf, ".html");
			is0h = FALSE;
		}

                if ( (gfp = fopen( tmpbuf, "r")) == (FILE *) NULL ) {
                        /* File is no longer there, keep going */
                        continue;
                }

		if ( (regp = regcomp( strlower(ip->extra))) == NULL ) {
			senderr2( "Search term regular expression error: ",
					ip->extra);
			break;
		}		
		if ( found_match( regp, gfp) ) {
			if ( http ) {
				if (firsttime == TRUE) {
					firsttime = FALSE;
					www_results( ip,
					"Directory Regular Expression Search",
						msgbuf);
				}
				www_cache_line( line1);
			}
			else
				send_cache_line( line1);

		}
		fclose( gfp);
	}
	if ( firsttime == TRUE) {
		sprintf( msgbuf, "%s <b>`%s'</b>.  %s", REGMISS,
			ip->extra, REGRES2);
		www_results( ip,
			"Directory Regular Expression Search", msgbuf);
	}
	cache_epilog( http);
	writelog( "Sent grep search", ip->selector);
}

static int
found_match( rp, fp)
struct regprog	*rp;
FILE	*fp;
{
	char linebuf[BUFSIZE];

	while ( fgets( linebuf, BUFSIZE, fp)) {
		strlower( linebuf);
		if ( regfind( rp, linebuf) )
			return TRUE;
	}
	return FALSE;
}

