/*
    Gn: A Server for the Internet Gopher Protocol(*).
    File: gn/sio.c
    Version 1.0
    
    Copyright (C) 1993  <by John Franks>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 1, or (at your option)
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    (*) Gopher is a registered trademark of the Univ. of Minn.
*/

#include <stdio.h>
#include <string.h>
#include "gn.h"

#define	EMPTY	(0)
#define USE1	(1)
#define USE2	(2)


static char	cacheline1[MAXLEN],
		cacheline2[MAXLEN];

static int	isgnlink,
		isremote;

static void	www_escape();


char *
get_input( sel)
char	*sel;
{
	if ( fgets( sel, PATHLEN, stdin) != NULL) {
		chop( sel);
		return ( sel);
	}
	return NULL;
}


void
send_text_line( line)
char	*line;
{
	if ( strchr( line, '\n') != NULL )
		chop( line);
	if ( http) {
		fputs( line, stdout);
		fputs( "\n", stdout);
	}
	else {
		if ( streq( line, "."))
		strcat( line, ".");
			/* Double a '.' if on line by itself */
		fputs( line, stdout);
		fputs( "\r\n", stdout);
	}
}

void
send_cache_line( line)
char	*line;
{
	char	*cp,
		*cp2;
	int	brackdepth = 0;

	Item    item;
	Cache_entry cache_entry;

	if ( strncmp( line, "http:", 5) == 0 )
		return;

	chop( line);
	if ( accesstype == DIRCHK) {
		parse_cache(line, &cache_entry);
		if ( !isremote && ( *(cache_entry.path) == '1')) {
			parse_selector( &item, cache_entry.path, FULL);
			if (!chkaccess( &item, NOLOG))  
				return;
		}
	}
	if ( *line == 'i') {  /* Delete everything in <> angle brackets */
		cp = cp2 = line;
		while ( *cp2 ) {
			if ( *cp2 == '<')
				brackdepth++;
			if ( brackdepth >0 )
				cp2++;
			else 
				*cp++ = *cp2++;
			if ( *(cp2 - 1) == '>')
				brackdepth--;
		}
		*cp = '\0';
	}
	fputs( line, stdout);
	fputs( "\r\n", stdout);
}

void
send_end()
{
	fputs( ".\r\n", stdout);
}



/*
 * read_cache() reads the cache file and puts a pointer to the primary
 * line in *l2.  A pointer to the secondary line (minus the leading tab)
 * is put in *l2.  If either doesn't exist NULL is put *l1 or *l2.
 * This function also sets the static variable 'isgnlink'.  The function
 * returns a pointer to the primary line.  If the item is a remote line
 * *l2 is made the empty string.
 */

char *
read_cache( l1, l2, fp)
char	**l1,
	**l2;
FILE	*fp;

{
	int	c;

	isgnlink = TRUE;
	isremote = FALSE;
	*l1 = cacheline1;
	*l2 = cacheline2;
	do {
		if ( fgets( cacheline1, MAXLEN, fp) == NULL) {
			*l1 = *l2 = NULL;
			return NULL;
		}
	} while ( *cacheline1 == '\t');

	chop( cacheline1);
	if ( ( c = fgetc( fp)) == '\t') {
		if ( fgets( cacheline2, MAXLEN, fp) == NULL) {
			*l2 = NULL;
			return cacheline1;
		}
		chop( cacheline2);
		if ( streq( cacheline2, "gopher_link\t")) {
			isgnlink = FALSE;
			isremote = TRUE;
			*cacheline2 = '\0';
		}
		if ( streq( cacheline2, "gn_link\t")) {
			isremote = TRUE;
			*cacheline2 = '\0';
		}
		return cacheline1;
	}
	else {
		ungetc( c, fp);
		*l2 = NULL;
		return cacheline1;
	}
}


void
www_cache_line( line)
char	*line;
{
	register char	*cp,
			*cp2;
	char		*prog,
			buf[PATHLEN];

	Item    	item;
	Cache_entry	cache_entry,
			*chp;
	char	out[PATHLEN];

	chp = &cache_entry;

	switch (*line) {
	case 'h':
		if ( strncmp( line, "http:", 5) == 0 ) {
			strcpy( out, line+5);
			send_text_line( out);
			return;
		}
		return;
	case 'i':
		strcpy(out, ++line);
		cp = strchr( out, '\t');
		*cp = '\0';
		send_text_line( out);
		return;
	}
	parse_cache(line, chp);
	if ( accesstype == DIRCHK) {
		if ( !isremote && ( *(chp->path) == '1')) {
			strcpy( buf, chp->path);
			parse_selector( &item, buf, FULL);
			if (!chkaccess( &item, NOLOG))  
				return;
		}
	}
	www_escape( chp->path);

	switch (*line) {
	case '8':
	case 'T':
		strcpy( buf, line);
		prog = ( *line == '8' ? "telnet" : "tn3270");
		if ( ( cp2 = strrchr( line, '\t')) == NULL) {
			senderr2( "Corrupt .cache file", "");
			exit(2);
		}
		*cp2++ = '\0';
		if ( ( cp = strrchr( line, '\t')) == NULL) {
			senderr2( "Corrupt .cache file", "");
			exit(2);
		}
		sprintf( out, "<li> <a href=\"%s://%s:%s\">%s</a>", 
			prog, cp, cp2, chp->name);
		break;
	default:

		if ( (strncmp( chp->path, "ftp", 3) == 0 ) &&
				((cp2 =  strchr(chp->path, ':')) != NULL) &&
				((cp =  strchr(cp2, '@')) != NULL) ) {
			*cp++ = '\0';
			cp2++;
			sprintf( out, "<li> <a href=\"ftp://%s%s\">%s</a>", 
				cp2, cp, chp->name);
			break;
		}
		if ( isgnlink) {
			sprintf(out,"<li> <a href=\"http://%s:%s/%s\">%s</a>", 
				chp->host, chp->port, chp->path, chp->name);
			break;
		}
		sprintf(out, "<li> <a href=\"gopher://%s:%s/%c%s\">%s</a>", 
			chp->host, chp->port, *line, chp->path, chp->name);

	}
	send_text_line( out);
}




/*
 * www_escape( path) replaces special chars in path with %XX escapes
 */

static void
www_escape( path)
char	*path;
{

	register char	*cp,
			*cp2;
	char	buf[BUFSIZE];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		switch (*cp) {
		case '"':
		case ' ':
			sprintf( cp2, "%%%X", (int) *cp);
			cp2 += 3;
			cp++;
			break;
		default:
			*cp2++ = *cp++;
		}
	}
	*cp2 = '\0';
	strcpy( path, buf);
}

/*
 * www_unescape( path) undoes what www_escape does
 */

void
www_unescape( path)
char	*path;
{

	register char	*cp,
			*cp2;
	int	val;

	char	minbuf[3],
		buf[SMALLLEN];

	cp = path;
	cp2 = buf;
	while ( *cp ) {
		if ( *cp == '%') {
			cp++;
			minbuf[0] = *cp++;
			minbuf[1] = *cp++;
			minbuf[3] = '\0';
			sscanf( minbuf, "%x", &val);
			*cp2++ = (char) val;
		}
		else
			*cp2++ = *cp++;
	}
	*cp2 = '\0';
	strcpy( path, buf);
}


